// -------------------------------------------
// gMini : a minimal OpenGL/GLUT application
// for 3D graphics.
// Copyright (C) 2006-2008 Tamy Boubekeur
// All rights reserved.
// -------------------------------------------

// -------------------------------------------
// Disclaimer: this code is dirty in the
// meaning that there is no attention paid to
// proper class attribute access, memory
// management or optimisation of any kind. It
// is designed for quick-and-dirty testing
// purpose.
// -------------------------------------------

#include "src/OpenGL.h"
#include <GLFW/glfw3.h>



#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cstdio>
#include <cstdlib>

#include <algorithm>
#include "src/Vec3.h"
#include "src/Camera.h"
#include "src/Mesh.h"


#include "src/imageLoader.h"

#include "src/mysparsematrix.h"
#include "src/generalEigenProblem.h"
#include "src/linearSystem.h"

using namespace std;



// window parameters
GLFWwindow *g_window = NULL;
int g_windowWidth = 1024;
int g_windowHeight = 768;




static Camera camera;
static bool mouseRotatePressed = false;
static bool mouseMovePressed = false;
static bool mouseZoomPressed = false;
static bool mouseIsPressed = false;
static int lastX=0, lastY=0, lastZoom=0;
static unsigned int FPS = 0;
static bool fullScreen = false;









Mesh mesh;

GLuint texture_handleIndex;
double textureScale = 1.0;
int visualizedEigenvalue = -1;

Eigen::Matrix< double, Eigen::Dynamic, Eigen::Dynamic > laplacianEigenvectors , laplacianEigenvalues;
Eigen::SparseMatrix< double > LaplacianMatrix;
Eigen::SparseMatrix< double > GradientMatrix;
Eigen::SparseMatrix< double > MassMatrix;
linearSystem HeatMethodSolver1;
linearSystem HeatMethodSolver2;
std::vector< double > scalarFunction;



double computeDotProductOfEigenVectors( unsigned int i , unsigned int j ) {
    double dotPr = 0.0;
    for (unsigned int v = 0; v < mesh.V.size (); v++) {
        dotPr += laplacianEigenvectors( v , i ) * MassMatrix.coeff(v,v) * laplacianEigenvectors( v , j );
    }
    return dotPr;
}

// THIS STEP IS KIND OF UNUSUAL... I THINK EIGEN DOES NOT DO IT ON PURPOSE, AND THAT EIGENVALUES ARE SOMEHOW INVOLVED HERE...
void renormalizeEigenVectors() {
    for( int ev = 0 ; ev < laplacianEigenvectors.cols() ; ++ev ) {
        double evNorm = sqrt(computeDotProductOfEigenVectors(ev,ev));
        for (unsigned int v = 0; v < mesh.V.size (); v++) {
            laplacianEigenvectors( v , ev ) /= evNorm;
        }
    }
}
void negateEigenValues() {
    for( unsigned int i = 0 ; i < laplacianEigenvalues.rows() ; ++i ) {
        laplacianEigenvalues(i,0) = - laplacianEigenvalues(i,0);
    }
}



void computeLaplacianEigenvectors( unsigned int nEigenvectors = 50 ) {
    // CAREFUL HERE ! YOU NEED TO ASK FOR THE EIGENVALUES OF -L,
    // BECAUSE THE LAPLACIAN HAS NEGATIVE EIGENVALUES AND EIGEN IMPLEMENTATION REQUIRE POSITIVE ONES:

    generalEigenProblem::extractSmallEigenVectors< double >( -1.0 * LaplacianMatrix , MassMatrix ,
                                                                  laplacianEigenvectors , laplacianEigenvalues ,
                                                                  nEigenvectors );
    std::cout << "Eigenvectors were extracted" << std::endl;

    renormalizeEigenVectors();
    negateEigenValues();

    for( int i = 0 ; i < 10 ; ++i ) {
        std::cout << "Eigenvalue " << i << " = " << laplacianEigenvalues(i,0) << std::endl;
        for( int j = 0 ; j < 10 ; ++j ) {
            std::cout << "   < eigVector " << i << " | eigVector " << j << " > = " << computeDotProductOfEigenVectors(i,j) << std::endl;
        }
    }
}






// ALREADY SEEN:
void computeBasisFunctionsGradients(Vec3 const & p0 , Vec3 const & p1, Vec3 const & p2 ,
                                    Vec3 & n , Vec3 & g0 , Vec3 & g1 , Vec3 & g2) {
    n = Vec3::cross( p1 - p0 , p2 - p0 );
    n.normalize();

    Vec3 r0 = Vec3::cross( n , p2 - p1 );
    g0 = r0 / Vec3::dot( p0 - p1 , r0 );

    Vec3 r1 = Vec3::cross( n , p0 - p2 );
    g1 = r1 / Vec3::dot( p1 - p2 , r1 );

    Vec3 r2 = Vec3::cross( n , p1 - p0 );
    g2 = r2 / Vec3::dot( p2 - p0 , r2 );
}

// ALREADY SEEN:
void computeGradientMatrix() {
    mySparseMatrix<double> myGradientMatrix( 3 * mesh.T.size() , mesh.V.size() );

    for( unsigned int t = 0 ; t < mesh.T.size() ; ++t ) {
        unsigned int v0 = mesh.T[t].v[0] , v1 = mesh.T[t].v[1] , v2 = mesh.T[t].v[2];
        Vec3 p0 = mesh.V[v0].p ,  p1 = mesh.V[v1].p ,  p2 = mesh.V[v2].p;
        Vec3 n , g0 , g1 , g2;
        computeBasisFunctionsGradients(p0 , p1 , p2 , n , g0 , g1 , g2);
        for(unsigned int coordonnee = 0 ; coordonnee < 3 ; ++coordonnee) {
            myGradientMatrix( 3*t + coordonnee , v0 ) = g0[coordonnee];
            myGradientMatrix( 3*t + coordonnee , v1 ) = g1[coordonnee];
            myGradientMatrix( 3*t + coordonnee , v2 ) = g2[coordonnee];
        }
    }

    GradientMatrix = myGradientMatrix.toEigenSparse< double >();
}







//-----------------------------------------    EXERCISES  :   --------------------------------------//

//void computeLaplacianMatrix() {
//    mySparseMatrix<double> myLaplacianMatrix( mesh.V.size() , mesh.V.size() );

//    for( unsigned int t = 0 ; t < mesh.T.size() ; ++t ) {
//        unsigned int v0 = mesh.T[t].v[0] , v1 = mesh.T[t].v[1] , v2 = mesh.T[t].v[2];
//        Vec3 p0 = mesh.V[v0].p ,  p1 = mesh.V[v1].p ,  p2 = mesh.V[v2].p;

//        //---------------------------------------------------//
//        //------------------   TO MODIFY  -------------------//
//        //---------------------------------------------------//
//        //cotanWeight(p1pOp2) += 1/2 * cotan(p1pOp2) :
//        double cotW0_by_2 = 0.0;
//        double cotW1_by_2 = 0.0;
//        double cotW2_by_2 = 0.0;
//        //---------------------------------------------------//
//        //---------------------------------------------------//
//        //---------------------------------------------------//




//        //---------------------------------------------------//
//        //-----------------   TO COMPLETE  ------------------//
//        //---------------------------------------------------//
//        myLaplacianMatrix(v1,v2) += cotW0_by_2;
//        myLaplacianMatrix(v1,v1) -= cotW0_by_2;
//        myLaplacianMatrix(v2,v1) += cotW0_by_2;
//        myLaplacianMatrix(v2,v2) -= cotW0_by_2;
//        //---------------------------------------------------//
//        //---------------------------------------------------//
//        //---------------------------------------------------//
//    }

//    LaplacianMatrix = myLaplacianMatrix.toEigenSparse< double >();
//}


void computeLaplacianMatrix() {
    mySparseMatrix<double> myLaplacianMatrix( mesh.V.size() , mesh.V.size() );

    for( unsigned int t = 0 ; t < mesh.T.size() ; ++t ) {
        unsigned int v0 = mesh.T[t].v[0] , v1 = mesh.T[t].v[1] , v2 = mesh.T[t].v[2];
        Vec3 p0 = mesh.V[v0].p ,  p1 = mesh.V[v1].p ,  p2 = mesh.V[v2].p;
        double p0p1_slength = (p1-p0).squareLength() , p1p2_slength = (p2-p1).squareLength() , p2p0_slength = (p0-p2).squareLength();
        double dot0 = Vec3::dot(p1-p0,p2-p0) , dot1 = Vec3::dot(p0-p1,p2-p1) , dot2 = Vec3::dot(p0-p2,p1-p2);

        //cotanWeight(p1pOp2) += 1/2 * cotan(p1pOp2)
        double cotW0_by_2 = dot0  / ( 2.0 * sqrt(p0p1_slength*p2p0_slength - dot0*dot0) );
        double cotW1_by_2 = dot1  / ( 2.0 * sqrt(p1p2_slength*p0p1_slength - dot1*dot1) );
        double cotW2_by_2 = dot2  / ( 2.0 * sqrt(p1p2_slength*p2p0_slength - dot2*dot2) );

        myLaplacianMatrix(v1,v2) += cotW0_by_2;     myLaplacianMatrix(v1,v1) -= cotW0_by_2;
        myLaplacianMatrix(v2,v1) += cotW0_by_2;     myLaplacianMatrix(v2,v2) -= cotW0_by_2;

        myLaplacianMatrix(v0,v2) += cotW1_by_2;     myLaplacianMatrix(v0,v0) -= cotW1_by_2;
        myLaplacianMatrix(v2,v0) += cotW1_by_2;     myLaplacianMatrix(v2,v2) -= cotW1_by_2;

        myLaplacianMatrix(v1,v0) += cotW2_by_2;     myLaplacianMatrix(v1,v1) -= cotW2_by_2;
        myLaplacianMatrix(v0,v1) += cotW2_by_2;     myLaplacianMatrix(v0,v0) -= cotW2_by_2;
    }

    LaplacianMatrix = myLaplacianMatrix.toEigenSparse< double >();
}






void computeMassMatrix() {
    mySparseMatrix<double> myMassMatrix( mesh.V.size() , mesh.V.size() );

    for( unsigned int t = 0 ; t < mesh.T.size() ; ++t ) {
        unsigned int v0 = mesh.T[t].v[0] , v1 = mesh.T[t].v[1] , v2 = mesh.T[t].v[2];
        Vec3 p0 = mesh.V[v0].p ,  p1 = mesh.V[v1].p ,  p2 = mesh.V[v2].p;
        double t_area_by_3 = Vec3::cross( p1 - p0 , p2 - p0 ).length() / 6.0;

        myMassMatrix(v0,v0) += t_area_by_3;
        myMassMatrix(v1,v1) += t_area_by_3;
        myMassMatrix(v2,v2) += t_area_by_3;
    }

    MassMatrix = myMassMatrix.toEigenSparse< double >();
}





void computeCommuteDistance( unsigned int sourceVertex , std::vector< double > & distanceToSource , unsigned int nEigenvaluesToSkip = 1 ) {
    // typically, on manifolds and using the Laplacian eigenmodes, we skip the first vector, which is constant and corresponds to eigenvalue 0
    // set appropriately, based on the problem you are considering

    distanceToSource.clear();
    distanceToSource.resize( mesh.V.size() , 0.0 );

    for( unsigned int i = nEigenvaluesToSkip ; i < laplacianEigenvalues.rows() ; ++i ) {
        double lambda_i = laplacianEigenvalues(i,0);
        if(lambda_i == 0.0) continue;

        double mu_i_sourceVertex = laplacianEigenvectors(sourceVertex,i);
        for( unsigned int v = 0 ; v < mesh.V.size() ; ++v ) {
            double mu_i_v = laplacianEigenvectors(v,i);

            //---------------------------------------------------//
            //-----------------   TO COMPLETE  ------------------//
            //---------------------------------------------------//
            distanceToSource[v] += 0.0;
            //---------------------------------------------------//
            //---------------------------------------------------//
            //---------------------------------------------------//
        }
    }
}



void computeBiharmonicDistance( unsigned int sourceVertex , std::vector< double > & distanceToSource , unsigned int nEigenvaluesToSkip = 1 ) {
    // typically, on manifolds and using the Laplacian eigenmodes, we skip the first vector, which is constant and corresponds to eigenvalue 0
    // set appropriately, based on the problem you are considering

    distanceToSource.clear();
    distanceToSource.resize( mesh.V.size() , 0.0 );

    //---------------------------------------------------//
    //-----------------   TO COMPLETE  ------------------//
    //---------------------------------------------------//
}



void computeDiffusionDistance( unsigned int sourceVertex , std::vector< double > & distanceToSource , double t ) {
    distanceToSource.clear();
    distanceToSource.resize( mesh.V.size() , 0.0 );


    //---------------------------------------------------//
    //-----------------   TO COMPLETE  ------------------//
    //---------------------------------------------------//
}

void computeHeatKernel( double t , std::vector< double > & heatKernel ) {
    heatKernel.clear();
    heatKernel.resize(mesh.V.size() , 0.0);


    //---------------------------------------------------//
    //-----------------   TO COMPLETE  ------------------//
    //---------------------------------------------------//
}


void precomputeSolversForTheHeatMethod( double t ) {
    HeatMethodSolver1.setDimensions( mesh.V.size() , mesh.V.size() );
    HeatMethodSolver1.setA_and_preprocess( MassMatrix - t * LaplacianMatrix );


    //---------------------------------------------------//
    //-----------------   TO COMPLETE  ------------------//
    //---------------------------------------------------//
    HeatMethodSolver2.setDimensions( 3 * mesh.T.size() + 1 , mesh.V.size() );
    //---------------------------------------------------//
    //---------------------------------------------------//
    //---------------------------------------------------//

    HeatMethodSolver2.preprocess();
}
void computeGeodesicDistanceUsingTheHeatMethod( unsigned int sourceIdx , std::vector< double > & distanceToSource ) {
    for( unsigned int v = 0 ; v < mesh.V.size() ; ++v ) {
        HeatMethodSolver1.b(v) = 0.0;
    }
    HeatMethodSolver1.b(sourceIdx) = MassMatrix.coeff(sourceIdx,sourceIdx);

    Eigen::VectorXd Ut;
    HeatMethodSolver1.solve(Ut);


    //---------------------------------------------------//
    //-----------------   TO COMPLETE  ------------------//
    //---------------------------------------------------//
}


void draw_initial() {
    mesh.draw();
}

void draw_exercise_1() {
    if(visualizedEigenvalue < 0  ||  visualizedEigenvalue >= laplacianEigenvectors.cols()) {
        draw_initial();
        return;
    }

    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D , texture_handleIndex);
    glColor3f(1,1,1);

    double normalizationFactor = 0.0;
    for (unsigned int i = 0; i < mesh.V.size (); i++) {
        normalizationFactor = std::max<double>(normalizationFactor , fabs( laplacianEigenvectors( i , visualizedEigenvalue ) ) );
    }

    glBegin (GL_TRIANGLES);
    for (unsigned int i = 0; i < mesh.T.size (); i++) {
        for (unsigned int j = 0; j < 3; j++) {
            const MeshVertex & vertex = mesh.V[mesh.T[i].v[j]];
            // put range [-1,1] to [0,1] in the texture coordinate (black line is at 0.5 in the texture)
            glTexCoord2f( 0.5f + 0.5f * laplacianEigenvectors( mesh.T[i].v[j] , visualizedEigenvalue ) / normalizationFactor, 0.f);
            glNormal3f (vertex.n[0], vertex.n[1], vertex.n[2]);
            glVertex3f (vertex.p[0], vertex.p[1], vertex.p[2]);
        }
    }
    glEnd ();

    glDisable(GL_TEXTURE_2D);
}


void draw_positive_scalar_function( ) {
    if( scalarFunction.size() != mesh.V.size() ) {
        draw_initial();
        return;
    }

    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D , texture_handleIndex);
    glColor3f(1,1,1);

    double normalizationFactor = 0.0;
    for (unsigned int i = 0; i < mesh.V.size (); i++) {
        normalizationFactor = std::max<double>(normalizationFactor , scalarFunction[i] );
    }

    glBegin (GL_TRIANGLES);
    for (unsigned int i = 0; i < mesh.T.size (); i++) {
        for (unsigned int j = 0; j < 3; j++) {
            const MeshVertex & vertex = mesh.V[mesh.T[i].v[j]];
            // put range [0,1] to [0.5,1] in the texture coordinate (black line is at 0.5 in the texture)
            glTexCoord2f( 0.5 + 0.5 * scalarFunction[ mesh.T[i].v[j] ] / normalizationFactor , 0.f );
            glNormal3f (vertex.n[0], vertex.n[1], vertex.n[2]);
            glVertex3f (vertex.p[0], vertex.p[1], vertex.p[2]);
        }
    }
    glEnd ();

    glDisable(GL_TEXTURE_2D);
}

void draw_exercise_2() {
    draw_positive_scalar_function( );
}









//---------------------------------   YOU DO NOT NEED TO CHANGE THE FOLLOWING CODE  --------------------------------//
void glVertex(Vec3 const & p) {
    glVertex3f(p[0] , p[1] , p[2]);
}
void glNormal(Vec3 const & p) {
    glNormal3f(p[0] , p[1] , p[2]);
}


/*
from your modelview:
|lx,ux,vx,px|     -- lx,ly,lz is your left-vector
|ly,uy,vy,py|     -- ux,uy,uz is your up-vector
|lz,uz,vz,pz|     -- vx,vy,vz is your view-vector
|0 ,0 ,0 ,1 |     -- px,py,pz is your translation
*/
Vec3 getRightVector() {
    float modelview[16];  glGetFloatv(GL_MODELVIEW_MATRIX , modelview);
    return Vec3( modelview[0] , modelview[4] , modelview[8] );
}
Vec3 getUpVector() {
    float modelview[16];  glGetFloatv(GL_MODELVIEW_MATRIX , modelview);
    return Vec3( modelview[1] , modelview[5] , modelview[9] );
}
Vec3 getViewVector() {
    float modelview[16];  glGetFloatv(GL_MODELVIEW_MATRIX , modelview);
    return Vec3( modelview[2] , modelview[6] , modelview[10] );
}





// ------------------------------------

void initLight () {
    GLfloat light_position1[4] = {22.0f, 16.0f, 50.0f, 0.0f};
    GLfloat direction1[3] = {-52.0f,-16.0f,-50.0f};
    GLfloat color1[4] = {1, 1,1, 1};
    GLfloat ambient[4] = {1, 1, 1, 1};

    glLightfv (GL_LIGHT1, GL_POSITION, light_position1);
    glLightfv (GL_LIGHT1, GL_SPOT_DIRECTION, direction1);
    glLightfv (GL_LIGHT1, GL_DIFFUSE, color1);
    glLightfv (GL_LIGHT1, GL_SPECULAR, color1);
    glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambient);
    glEnable (GL_LIGHT1);
    glEnable (GL_LIGHTING);
}

void init () {
    camera.resize (g_windowWidth, g_windowHeight);
    initLight ();
    glCullFace (GL_BACK);
    glEnable (GL_CULL_FACE);
    glDepthFunc (GL_LESS);
    glEnable (GL_DEPTH_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_COLOR_MATERIAL);
    glClearColor (0.9f, 0.9f, 0.9f, 1.0f);
}





void render () {
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glViewport(0, 0, g_windowWidth, g_windowHeight);
    glLoadIdentity ();
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    camera.apply ();

    glEnable(GL_DEPTH);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glDisable(GL_BLEND);
    glColor3f(0.4,0.4,0.8);

   // draw_initial();
    draw_exercise_1();
   // draw_exercise_2();
}








// Called each time a mouse button is pressed
void mouseButtonCallback(GLFWwindow *window, int button, int state, int mods)
{
    int width, height;
    glfwGetWindowSize(window, &width, &height);
    int x = lastX;
    int y = lastY;

    mouseIsPressed = state != GLFW_RELEASE;

    if( mods & GLFW_MOD_CONTROL    ) {

    }
    else {
        // moving the camera:
        if (state == GLFW_RELEASE) {
            mouseMovePressed = false;
            mouseRotatePressed = false;
            mouseZoomPressed = false;
        } else {
            if (button == GLFW_MOUSE_BUTTON_LEFT) {
                camera.beginRotate (x, y);
                mouseMovePressed = false;
                mouseRotatePressed = true;
                mouseZoomPressed = false;
            } else if (button == GLFW_MOUSE_BUTTON_RIGHT) {
                lastX = x;
                lastY = y;
                mouseMovePressed = true;
                mouseRotatePressed = false;
                mouseZoomPressed = false;
            } else if (button == GLFW_MOUSE_BUTTON_MIDDLE) {
                if (mouseZoomPressed == false) {
                    lastZoom = y;
                    mouseMovePressed = false;
                    mouseRotatePressed = false;
                    mouseZoomPressed = true;
                }
            }
        }
    }
}







// Called each time the mouse cursor moves
void cursorPosCallback(GLFWwindow *window, double xpos, double ypos)
{
    int width, height;
    glfwGetWindowSize(window, &width, &height);
    int x = xpos;
    int y = ypos;

    if (! mouseIsPressed) {
        lastX = x;
        lastY = y;
    }

    {
        // moving the camera:
        if (mouseRotatePressed == true) {
            camera.rotate (x, y);
        }
        else if (mouseMovePressed == true) {
            camera.move ((x-lastX)/static_cast<float>(g_windowWidth), (lastY-y)/static_cast<float>(g_windowHeight), 0.0);
            lastX = x;
            lastY = y;
        }
        else if (mouseZoomPressed == true) {
            camera.zoom (float (y-lastZoom)/g_windowHeight);
            lastZoom = y;
        }
    }
}





// Executed each time the window is resized. Adjust the aspect ratio and the rendering viewport to the current window.
void windowSizeCallback(GLFWwindow *window, int width, int height)
{
    g_windowWidth = width;
    g_windowHeight = height;
    camera.resize (width, height);
    glViewport(0, 0, (GLint)width, (GLint)height); // Dimension of the rendering region in the window
}










// Executed each time a key is entered.
void keyCallback(GLFWwindow *window, int key, int scancode, int action, int mods)
{
    if (action == GLFW_RELEASE) return;

    switch (key) {
    case GLFW_KEY_UP :
        textureScale *= 1.1;
        break;
    case GLFW_KEY_DOWN :
        textureScale /= 1.1;
        break;
    case GLFW_KEY_LEFT :
        visualizedEigenvalue--;
        if(visualizedEigenvalue < -1)
            visualizedEigenvalue = -1;
        break;
    case GLFW_KEY_RIGHT :
        visualizedEigenvalue++;
        if(visualizedEigenvalue >= laplacianEigenvectors.cols())
            visualizedEigenvalue = laplacianEigenvectors.cols() - 1;
        break;
    default:
        break;
    }
}







void initGLFW()
{
    // Initialize GLFW, the library responsible for window management
    if(!glfwInit()) {
        std::cerr << "ERROR: Failed to init GLFW" << std::endl;
        std::exit(EXIT_FAILURE);
    }

    // Before creating the window, set some option flags
    // WE DON'T CARE ABOUT MODERN OPENGL, WE'RE DOING GEOMETRY HERE
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 2);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
   // glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
   // glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_RESIZABLE, GL_TRUE);

    // Create the window
    g_window = glfwCreateWindow(g_windowWidth, g_windowHeight, "IG3DA - Practical", NULL, NULL);
    if(!g_window) {
        std::cerr << "ERROR: Failed to open window" << std::endl;
        glfwTerminate();
        std::exit(EXIT_FAILURE);
    }

    // Load the OpenGL context in the GLFW window using GLAD OpenGL wrangler
    glfwMakeContextCurrent(g_window);

    // not mandatory for all, but MacOS X
    glfwGetFramebufferSize(g_window, &g_windowWidth, &g_windowHeight);

    // Connect the callbacks for interactive control
    glfwSetWindowSizeCallback(g_window, windowSizeCallback);
    glfwSetKeyCallback(g_window, keyCallback);
    glfwSetCursorPosCallback(g_window, cursorPosCallback);
    glfwSetMouseButtonCallback(g_window, mouseButtonCallback);
}




void exitOnCriticalError(const std::string &message)
{
    std::cerr << "> [Critical error]" << message << std::endl;
    std::cerr << "> [Exit]" << std::endl;
    std::exit(EXIT_FAILURE);
}

void initOpenGL()
{
    // Load extensions for modern OpenGL
//    if(!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
//        exitOnCriticalError("[Failed to initialize OpenGL context]");

    if(!gladLoadGL())
        exitOnCriticalError("[Failed to initialize OpenGL context]");

    // glEnable(GL_DEBUG_OUTPUT);    // Modern error callback functionality
    // glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS); // For recovering the line where the error occurs, set a debugger breakpoint in DebugMessageCallback
    // glDebugMessageCallback(debugMessageCallback, 0); // Specifies the function to call when an error message is generated.

    glCullFace(GL_BACK); // Specifies the faces to cull (here the ones pointing away from the camera)
    glEnable(GL_CULL_FACE); // Enables face culling (based on the orientation defined by the CW/CCW enumeration).
    glDepthFunc(GL_LESS);   // Specify the depth test for the z-buffer
    glEnable(GL_DEPTH_TEST);      // Enable the z-buffer test in the rasterization
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f); // specify the background color, used any time the framebuffer is cleared



    unsigned char * imageTexture; unsigned int w , h;
    ppmLoader::load_ppm(imageTexture , w,h, "./img/heatMap.ppm");
    std::cout << w << "  " << h << std::endl;
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1 , &texture_handleIndex);
    glBindTexture(GL_TEXTURE_2D , texture_handleIndex);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexImage2D(GL_TEXTURE_2D , 0 , GL_RGB , w , h , 0 , GL_RGB , GL_UNSIGNED_BYTE , imageTexture);
    glDisable(GL_TEXTURE_2D);
}






int main(int argc, char **argv)
{
    initGLFW();                   // Windowing system
    initOpenGL();                 // OpenGL Context and shader pipeline
    init();

    mesh.loadOFF(argc == 2 ? argv[1] : "models/arma_lowres.off");


    computeLaplacianMatrix();
    computeMassMatrix();

    computeLaplacianEigenvectors( 50 ); // uncomment draw_exercise_1() dans la fonction draw() to visualize the eigenvectors of the laplacian.

    //computeCommuteDistance( 0 , scalarFunction);

    //computeBiharmonicDistance( 0 , scalarFunction);

   // computeDiffusionDistance(0,scalarFunction,1.2); // TRY : 0.05  ,  0.1  ,  0.2  ,  0.35  ,  0.5  ,  0.8  ,  1.2

    // computeHeatKernel(1 , scalarFunction); // TRY : 0.001 ,  0.05  ,  0.1  ,  0.2  ,  0.35  ,  0.5  ,  0.8  ,  1.2
    // NOTE THAT FOR SOME VALUES, YOU MIGHT NEED TO COMPUTE MORE EIGENVECTORS THAN JUST 50...

   // computeGradientMatrix();
  //  precomputeSolversForTheHeatMethod(0.1);
  //  computeGeodesicDistanceUsingTheHeatMethod( 0 , scalarFunction );



    while(!glfwWindowShouldClose(g_window)) {
        // update(static_cast<float>(glfwGetTime()));
        render();
        glfwSwapBuffers(g_window);
        glfwPollEvents();
    }

    std::cout << " > Quit" << std::endl;
    return EXIT_SUCCESS;
}


