#ifndef generalEigenProblem_H
#define generalEigenProblem_H



#include <Eigen/Core>
#include <Eigen/SparseCore>
#include <Eigen/SparseCholesky>


#include <MatOp/SparseGenMatProd.h>
#include <MatOp/SparseSymShiftSolve.h>
#include <GenEigsSolver.h>
#include <SymEigsShiftSolver.h>
#include <SymGEigsSolver.h>
#include <MatOp/SparseCholesky.h>


namespace generalEigenProblem {
using namespace Spectra;

// Computes the eigenvectors X_i of the form : A . X_i = mu_i . B . X_i , with mu_i small:
    template< class float_t >
    bool extractSmallEigenVectors( Eigen::SparseMatrix< float_t > const & A , Eigen::SparseMatrix< float_t > const & B ,
                                   Eigen::Matrix< float_t, Eigen::Dynamic, Eigen::Dynamic > & evectors ,
                                   Eigen::Matrix< float_t, Eigen::Dynamic, Eigen::Dynamic > & evalues ,
                                   unsigned int nVectorsToFind = 1 ) {

        SparseCholesky< float_t >  op( A );
        SparseGenMatProd< float_t >  Bop( B );
        SymGEigsSolver< float_t , LARGEST_MAGN, SparseGenMatProd< float_t >, SparseCholesky< float_t >, GEIGS_CHOLESKY>
                geigs(&Bop, &op, nVectorsToFind, std::max<unsigned int>(nVectorsToFind + 1 , 1.2 * nVectorsToFind));

        // Initialize and compute
        geigs.init();
        geigs.compute();
        // Retrieve results
        if(geigs.info() == SUCCESSFUL) {
            evectors = geigs.eigenvectors();
            evalues = geigs.eigenvalues();
            return true;
        }

        std::cout << "Eigenvalues WERE NOT found" << std::endl;
        return false;
    }


//// Computes the eigenvectors X_i of the form : A . X_i = mu_i  B . X_i , with mu_i small:
//template< class float_t >
//bool extractSmallEigenVectors( Eigen::SparseMatrix< float_t > const & A , Eigen::SparseMatrix< float_t > const & B ,
//                               Eigen::Matrix< float_t, Eigen::Dynamic, Eigen::Dynamic > & evectors ,
//                               Eigen::Matrix< float_t, Eigen::Dynamic, Eigen::Dynamic > & evalues ,
//                               unsigned int nVectorsToFind = 1 ) {

//    // 3 secs on the armadillo:
//    // a (small) regularization is needed if you use this, because it will find eigenvalue = 1/0 (not well defined...) otherwise:
//    float_t epsilon = 0.000000000001;
//    Eigen::SparseMatrix< float_t > diagonalRegularizationMatrix(A.rows(),A.rows());
//    {
//        std::vector< Eigen::Triplet< float_t > > triplets;
//        for( unsigned int r = 0 ; r < A.rows() ; ++r ) {
//            triplets.push_back( Eigen::Triplet< float_t >(r,r,epsilon) );
//        }
//        diagonalRegularizationMatrix.setFromTriplets( triplets.begin() , triplets.end() );
//    }

//    SparseCholesky< float_t >  op( A + diagonalRegularizationMatrix );
//    SparseGenMatProd< float_t >  Bop( B );
//    SymGEigsSolver< float_t , LARGEST_MAGN, SparseGenMatProd< float_t >, SparseCholesky< float_t >, GEIGS_CHOLESKY>
//            geigs(&Bop, &op, nVectorsToFind, std::max<unsigned int>(nVectorsToFind + 1 , 3 * nVectorsToFind));

//    // Initialize and compute
//    geigs.init();
//    geigs.compute();
//    // Retrieve results
//    if(geigs.info() == SUCCESSFUL) {
//        evectors = geigs.eigenvectors();
//        evalues = geigs.eigenvalues();

//        // don't forget to invert the eigenvalues (since you computed the inverse general eigenproblem):
//        for( unsigned int i = 0 ; i < evalues.rows() ; ++i ) {
//            evalues(i,0) = 1.0 / evalues(i,0);
//        }

//        return true;
//    }

//    std::cout << "Eigenvalues WERE NOT found" << std::endl;
//    return false;
//}


}

#endif // generalEigenProblem_H
